"""
Our Asynchronously Interchangeable Objects Client
=================================================

this portion is providing a client interface to manage asynchronously interchangeable oai objectz.

each oai object has a user definable values dictionary and can optionally have a file or folder attached to it.

an instance of the here implemented class :class:`~ae.oaio_client.OaioClient`, represents a client interface
to the oaio server, a django server `available as git repository <https://gitlab.com/ae-group/oaio_server>`_.
"""
import os
import shutil

from ast import literal_eval
from copy import deepcopy
from typing import Optional, Sequence, cast

import requests

from ae.base import (                                                                                   # type: ignore
    app_name_guess, defuse, os_device_id, os_path_isdir, os_path_isfile, os_path_join, os_path_relpath,
    os_user_name, read_file, write_file,
    ErrorMsgMixin)
from ae.cloud_storage import CshApiBase, csh_api_class                                                  # type: ignore
from ae.oaio_model import (                                                                             # type: ignore
    CREATE_ACCESS_RIGHT, DELETE_ACTION, FILES_DIR, FILES_VALUES_KEY, MAX_STAMP_DIFF, NO_ACCESS_RIGHT, OBJECTS_DIR,
    READ_ACCESS_RIGHT, REGISTER_ACTION, ROOT_VALUES_KEY, UPLOAD_ACTION, OLDEST_SYNC_STAMP,
    OaioCshIdType, OaioMapType, OaioIdType, OaioUserIdType, OaioStampType, OaioValuesType, OaioAccessRightType,
    OaiObject, extra_headers, now_stamp, object_dict, object_id, stamp_diff)
from ae.paths import Collector, normalize, placeholder_path                                             # type: ignore


__version__ = '0.3.16'


LAST_SYNC_STAMP_FILE = 'last_sync_stamp.txt'    #: file name to store the last synchronization timestamp


UserzAccessType = list[dict[OaioUserIdType, OaioAccessRightType]]   #: :meth:`OaioClient.userz_access` return value


class OaioClient(ErrorMsgMixin):
    """ interface to manage creations, updates and deletions of oaio objects of a user.

    .. note:: after creating an instance the :meth:`.synchronize` method has to be called at least once.

    """
    def __init__(self, host: str, credentials: dict[str, str],
                 app_id: str = "",
                 device_id: str = "",
                 csh_default_id: OaioCshIdType = 'Digi',
                 client_root: str = "{usr}/oaio_root/",
                 ):
        """ initialize a new client instance to connect to the oaio server.

        :param host:            oaio server host name/address and optional port number.
        :param credentials:     oaio server user identification credentials kwargs (dict with the keys
                                'username' and 'password').
        :param app_id:          optional id/name of the app from where the client is connecting from.
                                defaults to :attr:`~ae.core.AppBase.app_name` attribute value (if :mod:`ae.core`
                                got included in project) or (if not set) to the return value of the
                                function :func:`~ae.base.app_name_guess`.
        :param device_id:       id of the client device.
                                defaults to :data:`~ae.base.os_device_id`.
        :param csh_default_id:  id for the default cloud storage server to use.
                                defaults to 'Digi'.
        :param client_root:     path to the folder on the local device where the oaio info and files gets cached.
                                defaults to the placeholder path "{usr}/oaio_root/".
        """
        assert defuse(app_id) == app_id, f"{app_id=} contains invalid characters {set(defuse(app_id)) - set(app_id)=}"
        assert defuse(device_id) == device_id, f"invalid chr {set(defuse(device_id)) - set(device_id)=} in {device_id=}"

        super().__init__()

        self.base_url = 'http' + ('' if host[:9] in ('localhost', '127.0.0.1') else 's') + f'://{host}/api/'
        self.credentials = credentials
        self.user_name = credentials.get('username') or os_user_name()
        self.app_id = app_id or self.cae and self.cae.app_name or app_name_guess()
        self.device_id = device_id or os_device_id
        self.csh_default_id = csh_default_id
        self.client_root = normalize(client_root)

        self.session = requests.Session()
        self.session.headers.update(extra_headers(self.user_name, self.device_id, self.app_id))
        self.connected = False

        self._init_client_folders()
        self._last_sync_stamp = ""
        self.client_objectz: OaioMapType = {}

    def __del__(self):
        """ automatic logout and close of the http session. """
        if self.connected:
            self._request('post', 'logout/')
        if self.session:
            self.session.close()

    def _client_file_root(self, oaio_id: OaioIdType, values: OaioValuesType) -> str:
        """ determine the path on client/device where the file(s) attached to an oaio get stored.

        :param oaio_id:         id of the oaio.
        :param values:          oaio values (client_values if upload else server_values).
        :return:                path string.
        """
        if ROOT_VALUES_KEY in values:
            root_folder = normalize(values[ROOT_VALUES_KEY])
        else:
            root_folder = os_path_join(self.client_root, FILES_DIR, oaio_id)
        return root_folder

    def _client_server_file_api_paths(self, oai_obj: OaiObject) -> tuple[list[str], Optional[CshApiBase], str, str]:
        """ get attached files, the cloud storage api id, and the files root paths on client and cloud storage host. """
        values = oai_obj.client_values
        files = values.get(FILES_VALUES_KEY, [])
        csh_id = oai_obj.csh_id
        if not files or not csh_id or (csh_api := self._csh_api(csh_id)) is None:
            if files:
                self.error_message = f"cloud storage {csh_id=} empty/invalid; {oai_obj=}"   # pragma: no cover
            return [], None, "", ""

        oaio_id = oai_obj.oaio_id
        client_path = self._client_file_root(oaio_id, values)
        server_path = os_path_join(oaio_id, oai_obj.client_stamp)

        return files, csh_api, client_path, server_path

    def _csh_api(self, csh_id: OaioCshIdType) -> Optional[CshApiBase]:
        """ determine api and credentials of cloud storage host specified by its id from the web server db/config. """
        res = self._request('get', f'csh_args/{csh_id}')
        if self._res_err(res, f"oaio server error in {csh_id=}-args fetch"):                # pragma: no cover
            return None

        csh_kwargs = res.json()                         # type: ignore # self._res_err ensures that res is not None
        api_class = csh_api_class(csh_id)
        return api_class(**csh_kwargs)

    def _delete_client_object(self, oaio_id: OaioIdType) -> bool:
        """ delete object locally on client from file system and client_objectz dict.

        :param oaio_id:         id of the oaio.
        :return:                True if oaio got deleted successfully locally on client, else False.
        """
        path = os_path_join(self.client_root, OBJECTS_DIR, oaio_id)
        oai_obj = self.client_objectz.pop(oaio_id, None)
        if oai_obj and os_path_isfile(path):
            os.remove(path)
            return True

        self.error_message = f"delete client object integrity error: {path=} {oaio_id=} {oai_obj=}"
        return False

    def _download_object(self, oai_obj: OaiObject) -> bool:
        """ save oaio info and download optional attached files to local device/cache.

        :param oai_obj:         :class:`~ae.oaio_model.OaiObject` dataclass instance.
        :return:                True if object info and files got downloaded without errors, else False.
        """
        files, csh_api, client_path, server_path = self._client_server_file_api_paths(oai_obj)
        if files and csh_api is None:                                                       # pragma: no cover
            return False

        for file_path in files:
            assert csh_api is not None  # for mypy
            content = csh_api.deployed_file_content(os_path_join(server_path, file_path))
            if content is None:                                                             # pragma: no cover
                self.error_message = f"cloud storage download error: {csh_api.error_message=} {oai_obj=} {file_path=}"
                return False
            write_file(os_path_join(client_path, file_path), content, make_dirs=True)

        oai_obj.server_stamp = oai_obj.client_stamp
        oai_obj.server_values = deepcopy(oai_obj.client_values)

        self._save_client_object_info(oai_obj)

        return True

    def _folder_files(self, folder_path: str) -> Sequence[str]:
        """ collect all files under the specified root folder.

        :param folder_path:     root folder to collect files from (can contain path placeholders).
        :return:                list of file names (relative to the root/folder_path and without any path placeholders).
        """
        coll = Collector(main_app_name=self.app_id)
        coll.collect(folder_path, append=("**/*", "**/.*"), only_first_of=())
        return [os_path_relpath(_, folder_path) for _ in coll.files]

    def _init_client_folders(self):
        """ called on the first start of the client to create the folders and files under :attr:`.client_root`. """
        file_path = os_path_join(self.client_root, OBJECTS_DIR)
        if not os_path_isdir(file_path):
            os.makedirs(file_path)

        file_path = os_path_join(self.client_root, FILES_DIR)
        if not os_path_isdir(file_path):
            os.makedirs(file_path)

        file_path = os_path_join(self.client_root, LAST_SYNC_STAMP_FILE)        # add file with old self.last_sync_stamp
        if not os_path_isfile(file_path):                                       # to ensure sync on first app start
            write_file(file_path, OLDEST_SYNC_STAMP)

    def _load_client_object_info(self, oaio_id: OaioIdType, info_path: str = "") -> OaiObject:
        """ load actual client/local oaio info of the oaio specified by its id. """
        obj_lit = read_file(os_path_join(info_path or os_path_join(self.client_root, OBJECTS_DIR), oaio_id))
        obj_dict = literal_eval(obj_lit)
        return OaiObject(**obj_dict)

    def _load_client_objectz(self) -> OaioMapType:
        """ load actual client/local oaio infos of all local oai objects.

        :return:                mapping object with objectz stored on the local device.
        """
        objectz = {}
        info_path = os_path_join(self.client_root, OBJECTS_DIR)
        for oaio_id in os.listdir(info_path):
            oai_obj = self._load_client_object_info(oaio_id, info_path=info_path)
            objectz[oaio_id] = oai_obj
        return objectz

    def _load_server_objectz(self) -> OaioMapType:
        """ get from server all newly added and changed objectz of all the subscriptions of the current user.

        :return:                mapping object with objectz subscribed by the user from oaio server.
                                if the client is offline an empty mapping object.
        """
        objectz = {}
        res = self._request('get', 'oaio_stampz/')
        if res and res.ok:
            for oaio_dict in res.json():
                oai_obj = OaiObject(**oaio_dict)
                objectz[oai_obj.oaio_id] = oai_obj

        return objectz

    def _reconnect_check(self) -> bool:
        """ check if server got already connected and (re)connect if not.

        :return                 True if still connected or successfully reconnected
                                or False if offline, connection cannot be established, CSRF token not available or on
                                any other server login failure/error; in these cases check the message in the property
                                :attr:`~OaioClient.error_message` (inherited from :class:`~ae.base.ErrorMsgMixin`).
        """
        if self.connected:
            res = self._request('get', 'current_stamp/', _in_chk=True)
            if res and res.ok:
                client_stamp = now_stamp()
                server_stamp = res.json().get('current_stamp')
                if abs(stamp_diff(client_stamp, server_stamp)) > MAX_STAMP_DIFF:            # pragma: no cover
                    self.error_message = f"clocks are out of sync; {client_stamp=} {server_stamp=}"
                    self.connected = False
                    return False
                return True

            self.session.headers.pop('X-CSRFToken')                                         # pragma: no cover
            self.connected = False                                                          # pragma: no cover

        if not self.connected:
            res = self._request('post', 'login/', _in_chk=True, json=self.credentials)
            if not self._res_err(res, f"user '{self.user_name}' authentication error"):
                self.connected = True

        return self.connected

    def _request(self, method: str, slug: str, _in_chk: bool = False, **request_kwargs
                 ) -> Optional[requests.Response]:
        """ oaio server request checking CSRF for POST and automatic reconnect.

        :param method:          http request method (either 'get' or 'post').
        :param slug:            request slug added to base URL.
        :param _in_chk:         preventing endless recursion. only True for internal calls from
                                :meth:`~OaioClient._reconnect_check`, else False.
        :param request_kwargs:  extra kwargs passed onto the http request :paramref:`~_request.method` call.
        :return:                http response, or None if connection cannot be established (offline, authentication, ..)
                                or if any other error occurred (check self.error_message for details).
        """
        res = None
        url = self.base_url + slug
        try:
            if not _in_chk and not self._reconnect_check():
                return None

            hdr_token = self.session.headers.get('X-CSRFToken')     # only DELETE|PATCH|POST|PUT|... need a CSRF token
            ses_token = self.session.cookies.get('csrftoken')
            if method != 'get' and (not hdr_token or hdr_token != ses_token):
                if not hdr_token:  # works without this if (same: if hdr_token == ses_token), but with 10% more traffic?
                    res = self.session.get(url)                     # fetch new/changed CSRF token from oaio server
                    ses_token = self.session.cookies.get('csrftoken')
                    if err_msg := self._res_err(res, f"CSRF fetch error in {method=} with {url=}",
                                                failure="" if ses_token else ": empty token", add_err=False):
                        raise Exception(err_msg)                                            # pragma: no cover
                request_kwargs.setdefault('cookies', {'csrftoken': ses_token})
                self.session.headers['X-CSRFToken'] = cast(str, ses_token)  # w/o cast in types-requests 2.31.0.20240311
                self.session.headers['Referer'] = url               # needed only for https requests

            met = getattr(self.session, method)
            res = met(url, **request_kwargs)
            assert res, f"session {method=} call with {request_kwargs=} returned unexpected empty response {res=}"
            res.raise_for_status()
            self.error_message = ""
            return res

        except (requests.HTTPError, requests.ConnectionError, Exception) as exc:            # pragma: no cover
            if 'json' in request_kwargs and 'password' in request_kwargs['json']:   # hide password if in request_kwargs
                request_kwargs['json']['password'] = request_kwargs['json']['password'][:2] + "*" * 9
            self._res_err(res, f"{method=} {exc=} {url=} {_in_chk=} {request_kwargs=}")

        return None

    def _res_err(self, res: Optional[requests.Response], msg: str, failure: str = "", add_err: bool = True) -> str:
        """ web request response error checker, amending error message with status-code. """
        if not res or not res.ok or failure:
            msg += f"{failure};"
            if res:
                msg += f" {getattr(res, 'status_code', '?')=}"
                try:
                    msg += f" {res.json()=}"
                except (AttributeError, Exception):                                         # pragma: no cover
                    msg += f" {getattr(res, 'content', '¿')=}"
            if add_err:
                self.error_message = msg
            return msg
        return ""

    def _save_client_object_info(self, oai_obj: OaiObject):
        """ save oaio to local oaio info. """
        oaio_id = oai_obj.oaio_id
        assert oaio_id and oai_obj.client_stamp, f"cannot save oaio ({oai_obj}) locally with empty id or client_stamp"

        self.client_objectz[oaio_id] = oai_obj
        write_file(os_path_join(self.client_root, OBJECTS_DIR, oaio_id), repr(object_dict(oai_obj)))

    def _upload_object(self, oai_obj: OaiObject) -> bool:
        """ save local/client object and try to send changes to storage and oaio/web servers.

        :param oai_obj:         :class:`~ae.oaio_model.OaiObject` dataclass instance to update/upload/register.
        :return:                True if register/upload went well or False on failure or if servers are offline (for
                                details check the error_message attribute of this instance)..
        """
        files, csh_api, client_path, server_path = self._client_server_file_api_paths(oai_obj)
        if files and csh_api is None:
            return False                                                                    # pragma: no cover

        for file_path in files:
            assert csh_api is not None  # for mypy
            if not csh_api.deploy_file(os_path_join(server_path, file_path),
                                       source_path=os_path_join(client_path, file_path)):   # pragma: no cover
                self.error_message = f"'{file_path}' cloud storage upload error: {csh_api.error_message=} {oai_obj=}"
                return False

        action = UPLOAD_ACTION if oai_obj.server_stamp else REGISTER_ACTION
        res = self._request('post', f'{action}/', json=object_dict(oai_obj))
        if self._res_err(res, f"postponed {action} onto server of {oai_obj}"):
            return False                                                                    # pragma: no cover

        srv_dict = res.json()               # type: ignore # self._res_err ensures that res is not None
        assert srv_dict.get('oaio_id') == oai_obj.oaio_id, f"{srv_dict.get('oaio_id')=} not matching {oai_obj.oaio_id=}"

        oai_obj.server_stamp = oai_obj.client_stamp
        oai_obj.server_values = deepcopy(oai_obj.client_values)
        self._save_client_object_info(oai_obj)

        return True

    # public api of this client instance ##########################################################

    @property
    def last_sync_stamp(self) -> OaioStampType:
        """ timestamp of last synchronization with the oaio/storage servers.

        :getter:                timestamp of the last server synchronization.
        :setter:                any assigned error message will be accumulated to recent error messages.
                                pass an empty string to reset the error message.
        """
        if not self._last_sync_stamp:
            self._last_sync_stamp = read_file(os_path_join(self.client_root, LAST_SYNC_STAMP_FILE))
        return self._last_sync_stamp

    @last_sync_stamp.setter
    def last_sync_stamp(self, stamp: OaioStampType):
        write_file(os_path_join(self.client_root, LAST_SYNC_STAMP_FILE), stamp)
        self._last_sync_stamp = stamp

    def register_file(self, file_name: str, file_content: Optional[bytes] = None, root_path: str = "",
                      stamp: OaioStampType = "", csh_id: OaioCshIdType = "") -> Optional[OaiObject]:
        """ register new oaio file object.

        :param file_name:       name of the new file object to register.
        :param file_content:    file content if file not exists, pass bytes array to create the file.
        :param root_path:       root path on local-machine of the new file object to register.
                                using :attr:`.client_root` if not specified.
        :param stamp:           optional timestamp (created by :func:`~ae.oaio_model.now_stamp` if not specified).
        :param csh_id:          cloud storage server id or empty string to use default cloud storage server.
        :return:                new OaiObject instance
                                or None if either stamp or oaio_id are already registered (check self.error_message).
        :raises AssertionError: if file to register does not exist.
        """
        stamp = stamp or now_stamp()
        values = {FILES_VALUES_KEY: [file_name]}
        if root_path:
            values[ROOT_VALUES_KEY] = placeholder_path(root_path)
            oaio_id = ''
            file_path = normalize(os_path_join(root_path, file_name))
        else:
            oaio_id = object_id(user_name=self.user_name, device_id=self.device_id, app_id=self.app_id,
                                stamp=stamp, values=values)
            root_path = self._client_file_root(oaio_id, values)
            file_path = os_path_join(root_path, file_name)

        if file_content is None:
            assert os_path_isfile(file_path), f"file to register not found: {file_path}"
        else:
            write_file(file_path, file_content, make_dirs=True)

        return self.register_object(values, stamp=stamp, oaio_id=oaio_id, csh_id=csh_id)

    def register_folder(self, root_path: str, stamp: OaioStampType = "",
                        csh_id: OaioCshIdType = "") -> Optional[OaiObject]:
        """ register new oaio folder object.

        :param root_path:       root path on the local-machine/client of the new folder object to register.
        :param stamp:           optional timestamp (created by :func:`~ae.oaio_model.now_stamp` if not specified).
        :param csh_id:          cloud storage server id or empty string to use default cloud storage server.
        :return:                new OaiObject instance
                                or None if either stamp or oaio_id are already registered (check self.error_message).
        :raises AssertionError: if files to register does not exist.
        """
        root_path = normalize(root_path)
        values = {ROOT_VALUES_KEY: placeholder_path(root_path),
                  FILES_VALUES_KEY: self._folder_files(root_path)}

        return self.register_object(values, stamp=stamp, csh_id=csh_id)

    def register_object(self, values: OaioValuesType, stamp: OaioStampType = '',
                        oaio_id: OaioIdType = '', csh_id: OaioCshIdType = '') -> Optional[OaiObject]:
        """ register new oaio data object.

        :param values:          values data to register as a new oaio object.
        :param stamp:           optional timestamp (created by :func:`~ae.oaio_model.now_stamp` if not specified).
        :param oaio_id:         object id (will be created if not passed).
        :param csh_id:          cloud storage server id. using default cloud storage server if not specified or empty.
        :return:                new OaiObject instance.
                                or None if either stamp or oaio_id are already registered (check self.error_message).
        :raises AssertionError: on invalid argument values/types or if the oaio_id got already registered.
        """
        assert isinstance(values, dict), f"register_object(): values arg must be a dict; {values=}"    # OaioValuesType
        stamp = stamp or now_stamp()
        assert stamp > OLDEST_SYNC_STAMP, f"register_object(): too old {stamp=} specified; <= {OLDEST_SYNC_STAMP=}"

        oaio_id = oaio_id or object_id(user_name=self.user_name, device_id=self.device_id, app_id=self.app_id,
                                       stamp=stamp, values=values)
        assert oaio_id not in self.client_objectz, f"register_object(): {oaio_id=} got already registered"

        oai_obj = OaiObject(
            oaio_id=oaio_id,
            client_stamp=stamp,                     # could be changed by server on upload if conflicts with other stamp
            client_values=values,
            csh_id=csh_id or self.csh_default_id,
            username=self.user_name,
            device_id=self.device_id,
            app_id=self.app_id,
            csh_access_right=CREATE_ACCESS_RIGHT,   # registering owner has always all access rights
        )

        self._save_client_object_info(oai_obj)      # store new obj in local OBJECTS_DIR

        if not self.synchronize_with_server_if_online():
            return None                                                                     # pragma: no cover

        return oai_obj

    def synchronize_with_server_if_online(self) -> bool:
        """ synchronize local changes to server and any update/changes done on other clients from server to this client.

        .. hint:: if not connected to the oaio server then this method tries first to (re-)connect.

        :return:                False if the client is offline or on sync error, else True.
        """
        if not self._reconnect_check():
            return False

        soz = self._load_server_objectz()
        self.client_objectz = coz = self._load_client_objectz()
        error = False

        for cob in sorted(coz.values(), key=lambda _: _.client_stamp):
            if cob.client_stamp > cob.server_stamp:             # or cob.server_stamp == ''
                error = not self._upload_object(cob) or error   # _upload_object() extends self.error_message on error

        newer = [_ for _ in soz.values() if coz.get(_.oaio_id, OaiObject(_.oaio_id)).client_stamp < _.client_stamp]
        for sob in sorted(newer, key=lambda _: _.client_stamp):
            error = not self._download_object(sob) or error

        if not error:
            self.last_sync_stamp = now_stamp()

        return not error

    def unregister_object(self, oaio_id: OaioIdType, wipe_files: bool = False) -> str:
        """ unregister/delete oai object.

        :param oaio_id:         id of the oaio to unregister.
        :param wipe_files:      pass True to also remove/wipe all attached file(s) on the local machine.
        :return:                empty string if unregistering was successful, else error message on failure.
        """
        oai_obj = self.client_objectz.get(oaio_id)
        if oai_obj is None:
            self.error_message = f"client object to delete/unregister with id '{oaio_id}' not found"
            return self.error_message

        registered = oai_obj.server_stamp
        if registered:                          # if registered on server
            # noinspection PyTypeChecker
            res = self._request('post', f'{DELETE_ACTION}/', json=object_dict(oai_obj))     # PyCharm inspection Bug
            if self._res_err(res, f"object delete/unregister to server failed; {oai_obj=}"):
                return self.error_message

        if self._delete_client_object(oaio_id):
            if wipe_files:
                files, csh_api, client_path, server_path = self._client_server_file_api_paths(oai_obj)
                if files:
                    if registered:
                        err_msg = csh_api.delete_file_or_folder(server_path)    # type: ignore # csh_api is not None
                        if err_msg:
                            self.error_message = err_msg
                            return self.error_message
                    if os_path_isdir(client_path):
                        shutil.rmtree(client_path)

            self.synchronize_with_server_if_online()

        return self.error_message

    def unsubscribe(self, oaio_id: OaioIdType, user_name: OaioUserIdType) -> str:
        """ remove subscription of an oai object for a user.

        :param oaio_id:         id of the oaio to unsubscribe.
        :param user_name:       name of the subscribed user.
        :return:                empty string if subscription could be removed without error, else error message.
        """
        data = {'oaio_id': oaio_id, 'username': user_name, 'access_right': NO_ACCESS_RIGHT}
        res = self._request('post', 'subscribe/', json=data)
        if not self._res_err(res, f"unsubscribe of {user_name=} and {oaio_id=} failed"):
            if user_name == self.user_name:
                self._delete_client_object(oaio_id)

        return self.error_message

    def update_file(self, oaio_id: OaioIdType, file_name: str = "", file_content: Optional[bytes] = None,
                    root_path: str = "", stamp: OaioStampType = "") -> OaiObject:
        """ update oai file object locally.

        :param oaio_id:         id of the oai file object to update.
        :param file_name:       name (optionally with subdirectories path) of the attached file.
        :param file_content:    file content if file name not exists, pass bytes array to create the file.
        :param root_path:       root path on local-machine of the new file object to update.
                                using :attr:`.client_root` if not specified.
        :param stamp:           optional timestamp (using :func:`~ae.oaio_model.now_stamp` if not specified).
        :return:                the updated :class:`~ae.oaio_model.OaiObject` dataclass instance.
        :raises AssertionError: if file to update does not exist.
        """
        values = deepcopy(self.client_objectz[oaio_id].client_values)
        if file_name:
            values[FILES_VALUES_KEY] = [file_name]
        else:
            file_name = values[FILES_VALUES_KEY][0]
        if root_path:
            values[ROOT_VALUES_KEY] = placeholder_path(root_path)
        elif ROOT_VALUES_KEY in values:
            root_path = values[ROOT_VALUES_KEY]
        else:
            root_path = self._client_file_root(oaio_id, values)

        file_path = normalize(os_path_join(root_path, file_name))

        if file_content is None:
            assert os_path_isfile(file_path), f"file to update does not exist in {file_path=}"
        else:
            write_file(file_path, file_content, make_dirs=True)

        return self.update_object(oaio_id, values, stamp=stamp)

    def update_folder(self, oaio_id: OaioIdType, root_path: str = "", stamp: OaioStampType = "") -> OaiObject:
        """ update oai folder object locally.

        :param oaio_id:         id of the oai folder object to update.
        :param root_path:       new root path on local-machine of the new folder object to update.
                                using current root path if not specified.
        :param stamp:           optional timestamp (using :func:`~ae.oaio_model.now_stamp` if not specified).
        :return:                the updated :class:`~ae.oaio_model.OaiObject` dataclass instance.
        :raises AssertionError: if same file is in add_files and in removed_files,
                                or if one of the files to update does not exist.
        """
        values = deepcopy(self.client_objectz[oaio_id].client_values)
        if root_path:
            values[ROOT_VALUES_KEY] = placeholder_path(root_path)

        client_root = self._client_file_root(oaio_id, values)
        values[FILES_VALUES_KEY] = self._folder_files(client_root)

        return self.update_object(oaio_id, values, stamp=stamp)

    def update_object(self, oaio_id: OaioIdType, values: OaioValuesType, stamp: OaioStampType = "", reset: bool = False
                      ) -> OaiObject:
        """ update oaio data object locally.

        :param oaio_id:         id of a registered oaio to update.
        :param values:          values to update within the oaio.
        :param stamp:           optional timestamp (using :func:`~ae.oaio_model.now_stamp` if not specified).
        :param reset:           pass True to reset the values before they get updated with the values
                                specified in :paramref:`~update_object.values`.
        :return:                the updated :class:`~ae.oaio_model.OaiObject` dataclass instance.
        :raises AssertionError: on invalid argument values/types or if the oaio_id did not get registered.
        """
        assert oaio_id in self.client_objectz, f"update_object(): oaio object with id '{oaio_id}' not registered"
        oai_obj = self.client_objectz[oaio_id]
        assert oai_obj.client_stamp, f"update_object(): oai object {oaio_id=} has empty client_stamp; {oai_obj=}"

        if stamp:
            assert stamp > OLDEST_SYNC_STAMP, f"update_object(): too old stamp {stamp} specified; <={OLDEST_SYNC_STAMP}"
        else:                                                                               # pragma: no cover
            stamp = now_stamp()

        assert stamp > oai_obj.client_stamp, f"update_object(): got too old stamp {stamp}; > {oai_obj.client_stamp}"
        oai_obj.client_stamp = stamp

        if reset:
            oai_obj.client_values.clear()
        oai_obj.client_values.update(values)

        self._save_client_object_info(oai_obj)
        self.synchronize_with_server_if_online()

        return oai_obj

    def upsert_subscription(self, oaio_id: OaioIdType, user_name: OaioUserIdType,
                            access_right: OaioAccessRightType = READ_ACCESS_RIGHT) -> int:
        """ add/delete subscriber of an oai object or update the access right of an existing user subscription.

        :param oaio_id:         id of the oaio to subscribe to or to unsubscribe from.
        :param user_name:       name of the un-/subscribing user.
        :param access_right:    either :data:`~ae.oaio_model.NO_ACCESS_RIGHT` to unsubscribe user
                                or a user access right (one of :data:`~ae.oaio_model.ACCESS_RIGHTS`) to subscribe
                                the oaio specified by :paramref:`~upsert_subscription.oaio_id`
                                to a user specified by :paramref:`~upsert_subscription.user_name`.
                                e.g. to grant update rights pass the value :data:`~ae.oaio_model.UPDATE_ACCESS_RIGHT`.
                                if not specified then the value :data:`~ae.oaio_model.READ_ACCESS_RIGHT` will be used.
        :return:                the primary key integer value (Pubz.Pid) of the added/updated Pubz subscription record
                                or 0/zero if an error occurred.
        """
        data = {'oaio_id': oaio_id, 'username': user_name, 'access_right': access_right}
        res = self._request('post', 'subscribe/', json=data)    # json-kwarg implies content-type 'application/json' hdr
        if self._res_err(res, f"subscription of {user_name=} for {oaio_id=} failed"):
            return 0
        return res.json().get('Pid', 0)     # type: ignore # self._res_err ensures that res is not None

    def userz_access(self, oaio_id: OaioIdType) -> UserzAccessType:
        """ determine a list of all the registered users and their access right to the specified oaio.

        :param oaio_id:         id of the oaio determine users and access rights for
                                or an empty or unregistered oaio id if the oaio is not yet registered.
        :return:                list of dicts with the keys 'username' and 'access_right', or empty list on error. the
                                returned list is sorted by the access_right (create, delete, update, read) and username.
                                if an empty/unregistered oaio id got specified all registered userz will be returned
                                with the :data:`~ae.oaio_model.NO_ACCESS_RIGHT` access right.
        """
        res = self._request('get', f'user_subz/{oaio_id or "_not__existing__oaio_id_"}')
        warn_msg = self._res_err(res, f"Warning: OaioClient.userz_access() error; {oaio_id=}", add_err=False)
        if warn_msg:
            self.vpo(warn_msg)
            return []

        assert res, "for mypy - not seeing that self._res_err is checking res and returning empty string if not None"
        return sorted(res.json(), key=lambda _: _['access_right'] + _['username'])

    def wipe_user(self, user_name: OaioUserIdType) -> str:
        """ wipe the specified user including all its objects, subscriptions and log entries.

        :param user_name:       name of the subscribing user to be wiped from the database.
        :return:                empty string if user including its subscription and log entries could be removed,
                                else the error message.
        """
        res = self._request('post', 'wipe_user/', json={'Uid': user_name})
        self._res_err(res, f"wipe of user '{user_name}' failed")
        return self.error_message



""" unit/integration tests of the oaio client package. """
import os
import shutil
from typing import Optional
from unittest.mock import MagicMock, patch

import pytest
import requests

from ae.base import (
    TESTS_FOLDER, app_name_guess, load_dotenvs, os_device_id, os_path_abspath, os_path_dirname, os_path_isdir,
    os_path_isfile, os_path_join, os_user_name, read_file, write_file)
from ae.cloud_storage import CshApiBase, DigiApi
from ae.oaio_model import (
    CREATE_ACCESS_RIGHT, DELETE_ACTION, DOWNLOAD_ACTION, FILES_DIR, FILES_VALUES_KEY, HTTP_HEADER_USR_ID, OBJECTS_DIR,
    REGISTER_ACTION,
    ROOT_VALUES_KEY, UPDATE_ACCESS_RIGHT, OaiObject, UPLOAD_ACTION, now_stamp, object_dict)
from ae.paths import normalize, placeholder_path

from ae.oaio_client import LAST_SYNC_STAMP_FILE, OaioClient


load_dotenvs()


class TestInstantiation:
    """ client instantiation tests running on local machine and git/CI host, independent of mocked and live servers. """
    def test_args(self):
        cre_args = {'cre_key': 'cre_val'}
        tst_root = os_path_join(TESTS_FOLDER, 'tst')
        loc_path = os_path_join(tst_root, 'root/path')

        try:
            ocl = OaioClient("hOsT", cre_args, app_id='app_id', device_id='dev_id', csh_default_id='csh_id',
                             client_root=loc_path)

            assert "hOsT" in ocl.base_url
            assert ocl.credentials == cre_args
            assert ocl.app_id == 'app_id'
            assert ocl.device_id == 'dev_id'
            assert ocl.csh_default_id == 'csh_id'
            assert ocl.client_root == normalize(loc_path)

            assert os_path_isdir(os_path_join(loc_path, OBJECTS_DIR))
            assert os_path_isdir(os_path_join(loc_path, FILES_DIR))
            assert os_path_isfile(os_path_join(loc_path, LAST_SYNC_STAMP_FILE))

        finally:
            if os_path_isdir(tst_root):
                shutil.rmtree(tst_root)

    def test_args_defaults(self):
        ocl = OaioClient("", {})

        assert ocl.user_name == os_user_name()
        assert ocl.app_id == app_name_guess()
        assert ocl.device_id == os_device_id
        assert ocl.csh_default_id == 'Digi'
        assert ocl.client_root == normalize("{usr}/oaio_root/")

    def test_invalid_host_instantiation_args(self):
        ocl = OaioClient("", {})

        assert ocl
        assert not ocl.synchronize_with_server_if_online()


# ******   test environment for mocked unit and live integration tests   ******

other_user = 'tst_oth_usr'      # other user, for subscription sharing tests


def _csh_file_content(client_obj: OaioClient, oai_obj: OaiObject, file_count: int = 1) -> dict[str, bytes]:
    """ download content of attached file(s) from mocked or live cloud storage host. """
    # noinspection PyProtectedMember
    files, csh_api, client_path, server_path = client_obj._client_server_file_api_paths(oai_obj)
    assert csh_api
    assert len(files) == file_count

    return {fil: csh_api.deployed_file_content(os_path_join(server_path, fil)) for fil in files}


def _csh_file_remove(client_obj: OaioClient, oai_obj: OaiObject, file_count: int = 1) -> str:
    """ delete file content from mocked or live cloud storage host. """
    # noinspection PyProtectedMember
    files, csh_api, client_path, server_path = client_obj._client_server_file_api_paths(oai_obj)
    err_msg = client_obj.error_message
    if not csh_api:
        err_msg += "cloud storage host api could not be determined"
    if len(files) != file_count:
        err_msg += f"files count mismatch: {len(files)} != {file_count}"

    for fil in files:
        err_msg += csh_api.delete_file_or_folder(os_path_join(server_path, fil))
    err_msg += csh_api.delete_file_or_folder(os_path_dirname(server_path))  # also remove the folder named as oaio_id

    return err_msg


# ******   test environment for mocked unit tests   *******************************


EMPTY_FOLDER = "FOLDER HAS NO CONTENT: simulated directory left to allow separate deletion of it in tests"
CSH_API_ID = 'CshMock'          # referring to mock cloud storage class :class:`CshMockApi`
deployed_file_contents = {}     # keep/publish test files in mocked cloud storage for assertions

def _csh_mock_cleared_errors():
    err_msg = ""
    if not all(content == EMPTY_FOLDER for _file, content in deployed_file_contents.items()):
        files_left = {fi: co for fi, co in deployed_file_contents.items() if co != EMPTY_FOLDER}
        err_msg += f"_csh_mock_cleared_errors() found deployed and uncleared test files: {files_left}"
    return err_msg

class CshMockApi(CshApiBase):
    """ simulating dummy cloud storage server - available via :func:`csh_api_class` with id=='CshMock' """
    def __init__(self, **csh_args):
        super().__init__(**csh_args)
        self.csh_args = csh_args
        self.deployed_file_contents = deployed_file_contents

    def deployed_file_content(self, file_path: str) -> Optional[bytes]:
        ret = self.deployed_file_contents.get(file_path)
        if ret is None:
            self.error_message = f"CshMock-content-error: {file_path} not in {self.deployed_file_contents}"
        return ret

    def deploy_file(self, file_path: str, source_path: str = '') -> str:
        path = source_path or file_path
        if not os_path_isfile(path):
            self.error_message = f"CshMock-deploy-error: file '{path}' not found; id='{file_path}' src='{source_path}'"
            return ""

        content = read_file(path, extra_mode='b')
        self.deployed_file_contents[file_path] = content
        return file_path

    def delete_file_or_folder(self, file_path: str) -> str:
        files = self.deployed_file_contents
        found = files.pop(file_path, None) is not None
        if not found:
            for nam in files.copy().keys():
                if nam.startswith(file_path + '/'):
                    files.pop(nam)
                    found = True
            if not found:
                self.error_message = f"CshMock-delete-error {file_path} not in {files}"

        if found and len(left_parts := file_path.split('/')) > 2:
            dir_path = '/'.join(left_parts[:-1]) + '/'
            files[dir_path] = EMPTY_FOLDER

        return self.error_message

registered_oai_objectz: dict[str, OaiObject] = {}

def _mocked_get_or_post_success_response(url: str, **kwargs) -> requests.Response:
    """ requests.Session().get/.post method response mocks, simulating oaio server. """
    mock_res = MagicMock()
    mock_res.ok = True
    mock_res.status_code = 200
    mock_res.raise_for_status.return_value = None

    act_slug = url.split('/')[-2]
    if act_slug == 'current_stamp':
        mock_res.json.return_value = {'current_stamp': now_stamp()}
    elif act_slug == 'csh_args':
        mock_res.json.return_value = {}
    elif act_slug == REGISTER_ACTION:
        oai_obj = OaiObject(**kwargs['json'])
        registered_oai_objectz[oai_obj.oaio_id] = oai_obj
        mock_res.json.return_value = kwargs['json']
    elif act_slug == UPLOAD_ACTION:
        oai_obj = OaiObject(**kwargs['json'])
        registered_oai_objectz[oai_obj.oaio_id] = oai_obj
        mock_res.json.return_value = kwargs['json']
    elif act_slug == DELETE_ACTION:
        oai_obj = OaiObject(**kwargs['json'])
        registered_oai_objectz.pop(oai_obj.oaio_id)
        mock_res.json.return_value = kwargs['json']
    elif act_slug == 'oaio_stampz':
        mock_res.json.return_value = sorted((object_dict(_) for _ in registered_oai_objectz.values()),
                                            key=lambda _: _['client_stamp'])
    elif act_slug == 'subscribe':
        pid = 369963
        mock_res.json.return_value = {'Pid': pid, 'message': f"Pid:{pid} PUid/POid:{kwargs['json']}"}
    elif act_slug == 'user_subz':
        mock_res.json.return_value = [{'username': 'mocked_username', 'access_right': CREATE_ACCESS_RIGHT}]
    elif act_slug == 'wipe_user':
        mock_res.json.return_value = {'message': f"mocked success for wipe user {kwargs}"}
    elif 'json' in kwargs:
        mock_res.json.return_value = kwargs['json']
    else:                                                   # action slugs: DOWNLOAD_ACTION, 'login', 'logout'
        mock_res.json.return_value = {'message': f"mocked success for url {url} and {kwargs=}"}

    return mock_res

def _mocked_get_or_post_failure_response(url: str, **kwargs) -> requests.Response:
    """ requests.Session().get/.post method response mocks, returning error """
    mock_res = MagicMock()
    mock_res.ok = False
    mock_res.status_code = 489
    mock_res.json.return_value = {'e_r_r': f"mocked error for url {url} and {kwargs=}"}
    return mock_res

def user_client(usr_name: str):
    """ provide OaioClient instances simulating the specified user, connected to mocked oaio and cloud storage servers.

    the yield value is the OaioClient instance of the primary user.
    used as fixture via yield from for username parameterized unit tests.
    running also in CI tests on git repo host.

    referred and mocked `response` instance attributes and methods: ok, status_code, json(), raise_for_status()
    """
    root_path = os_path_join(TESTS_FOLDER, f'tst_rut_{usr_name}')
    with patch('ae.oaio_client.requests') as mock_req:
        mock_ses_cls = MagicMock()
        mock_req.Session = mock_ses_cls
        mock_ses_ins = mock_ses_cls.return_value
        mock_ses_ins.get.side_effect = _mocked_get_or_post_success_response
        mock_ses_ins.post.side_effect = _mocked_get_or_post_success_response

        ocl = OaioClient('any_host:8769', {'username': usr_name, 'password': 'any_password'},
                         app_id='aId', device_id='dId',
                         csh_default_id=CSH_API_ID,  # -> mocked class CshMockApi in this module
                         client_root=root_path,
                         )
        ocl.o_synchronize_with_server_if_online = ocl.synchronize_with_server_if_online
        ocl.synchronize_with_server_if_online = lambda: True  # monkey patch to allow manual server sync in tests

        assert ocl.error_message == ""
        assert os_path_isdir(root_path)

        yield ocl

        if os_path_isdir(root_path):
            shutil.rmtree(root_path)
        registered_oai_objectz.clear()
        deployed_file_contents.clear()

@pytest.fixture
def mocked_client(request):
    """ OaioClient instance for user uId """
    yield from user_client('uId')

@pytest.fixture
def other_client(mocked_client):
    """ OaioClient instance for user adaUsr """
    yield from user_client('adaUsr')


class TestOaioClientWithMockedServers:
    def test_mocked_client_initialization(self, mocked_client):
        assert mocked_client
        assert not mocked_client.error_message
        assert mocked_client.csh_default_id == CSH_API_ID
        assert not deployed_file_contents
        assert len(mocked_client.client_objectz) == 0

    def test_last_sync_stamp(self, mocked_client):
        assert mocked_client.last_sync_stamp
        stamp = now_stamp()
        mocked_client.last_sync_stamp = stamp
        assert mocked_client.last_sync_stamp == stamp

    def test_register_file(self, mocked_client):
        oai_obj = None
        stamp = now_stamp()
        dir_path = ""
        fil_name = 'tst_fil.xxx'
        fil_content = b"test file \x00 binary content \x01\x02\x03\xff"

        assert mocked_client.o_synchronize_with_server_if_online()
        local_object_count = len(mocked_client.client_objectz)

        try:
            oai_obj = mocked_client.register_file(fil_name, file_content=fil_content, stamp=stamp)

            oaio_id = oai_obj.oaio_id
            dir_path = mocked_client._client_file_root(oaio_id, oai_obj.client_values)
            assert mocked_client.error_message == ""
            assert oai_obj is not None
            assert oaio_id
            assert oai_obj.csh_id

            assert ROOT_VALUES_KEY not in oai_obj.client_values
            assert FILES_VALUES_KEY in oai_obj.client_values
            assert len(oai_obj.client_values[FILES_VALUES_KEY]) == 1
            assert oai_obj.client_values[FILES_VALUES_KEY][0] == fil_name
            assert os_path_isfile(os_path_join(dir_path, fil_name))
            assert not oai_obj.server_values
            assert oai_obj.client_stamp == stamp
            assert not oai_obj.server_stamp
            assert oai_obj.csh_access_right == CREATE_ACCESS_RIGHT

            assert len(mocked_client.client_objectz) == local_object_count + 1
            assert oaio_id in mocked_client.client_objectz
            assert mocked_client.client_objectz[oaio_id] == oai_obj

            assert mocked_client.o_synchronize_with_server_if_online()
            assert _csh_file_content(mocked_client, oai_obj) == {fil_name: fil_content}

        finally:
            if os_path_isdir(dir_path):
                shutil.rmtree(dir_path)
            if oai_obj:
                assert _csh_file_remove(mocked_client, oai_obj) == "", "ERR:" + mocked_client.error_message
                assert _csh_mock_cleared_errors() == ""
                assert mocked_client.unregister_object(oai_obj.oaio_id) == "", "ERR:" + mocked_client.error_message
                assert len(mocked_client.client_objectz) == local_object_count, "ERR:" + mocked_client.error_message

    def test_register_file_with_root(self, mocked_client):
        oai_obj = None
        stamp = now_stamp()
        dir_path = os_path_abspath(os_path_join(TESTS_FOLDER, 'regFileWithRootDir'))
        fil_name = 'tst_fil.tst'
        fil_path = os_path_join(dir_path, fil_name)
        fil_content = b"test file \x00 having binary content \x01\x02\x03\xff"

        assert mocked_client.o_synchronize_with_server_if_online()
        local_object_count = len(mocked_client.client_objectz)

        try:
            write_file(fil_path, fil_content, make_dirs=True)

            oai_obj = mocked_client.register_file(fil_name, root_path=dir_path, stamp=stamp)
            assert mocked_client.error_message == ""
            assert oai_obj is not None

            assert oai_obj.oaio_id
            assert oai_obj.csh_id
            assert ROOT_VALUES_KEY in oai_obj.client_values
            assert normalize(oai_obj.client_values[ROOT_VALUES_KEY]) == dir_path
            assert FILES_VALUES_KEY in oai_obj.client_values
            assert len(oai_obj.client_values[FILES_VALUES_KEY]) == 1
            assert oai_obj.client_values[FILES_VALUES_KEY][0] == fil_name
            assert not oai_obj.server_values
            assert oai_obj.client_stamp == stamp
            assert not oai_obj.server_stamp
            assert oai_obj.csh_access_right == CREATE_ACCESS_RIGHT

            assert len(mocked_client.client_objectz) == local_object_count + 1
            assert oai_obj.oaio_id in mocked_client.client_objectz
            assert mocked_client.client_objectz[oai_obj.oaio_id] == oai_obj

            assert os_path_isfile(fil_path)

            assert mocked_client.o_synchronize_with_server_if_online()
            assert _csh_file_content(mocked_client, oai_obj) == {fil_name: fil_content}

        finally:
            if os_path_isdir(dir_path):
                shutil.rmtree(dir_path)
            if oai_obj:
                assert _csh_file_remove(mocked_client, oai_obj) == "", "ERR:" + mocked_client.error_message
                assert _csh_mock_cleared_errors() == ""
                assert mocked_client.unregister_object(oai_obj.oaio_id) == "", "ERR:" + mocked_client.error_message
                assert len(mocked_client.client_objectz) == local_object_count

    def test_register_folder_with_root(self, mocked_client):
        stamp = now_stamp()
        # noinspection PyTypeChecker
        root_path = os_path_abspath(os_path_join(mocked_client.client_root, FILES_DIR))
        sub_dirs = ['sub_dir1', 'sub_dir3/sub_sub_dir']
        file_paths = ['tst_fil.tst',
                      os_path_join(sub_dirs[0], 'tst_sub1_fil.tst'),
                      os_path_join(sub_dirs[1], 'tst_sub3_sub_fil.tst'),
                      ]
        fil_content = b"test file \x00 binary content \x01\x02\x03\xff"

        assert mocked_client.o_synchronize_with_server_if_online()
        local_object_count = len(mocked_client.client_objectz)
        csh_api = mocked_client._csh_api(mocked_client.csh_default_id)
        oaio_id = ""

        try:
            for fil_nam in file_paths:
                write_file(os_path_join(root_path, fil_nam), fil_content + fil_nam.encode(), make_dirs=True)

            oai_obj = mocked_client.register_folder(root_path, stamp=stamp)

            assert oai_obj is not None
            oaio_id = oai_obj.oaio_id
            assert oaio_id
            assert mocked_client.error_message == ""

            svr_path = os_path_join(oaio_id, stamp)
            assert oai_obj.csh_id
            assert ROOT_VALUES_KEY in oai_obj.client_values
            assert normalize(oai_obj.client_values[ROOT_VALUES_KEY]) == normalize(root_path)
            assert FILES_VALUES_KEY in oai_obj.client_values
            assert len(oai_obj.client_values[FILES_VALUES_KEY]) == len(file_paths)
            for fil_nam in file_paths:
                assert fil_nam in oai_obj.client_values[FILES_VALUES_KEY]
                assert os_path_isfile(normalize(os_path_join(root_path, fil_nam)))
            assert not oai_obj.server_values
            assert oai_obj.client_stamp == stamp
            assert not oai_obj.server_stamp
            assert oai_obj.csh_access_right == CREATE_ACCESS_RIGHT

            assert len(mocked_client.client_objectz) == local_object_count + 1
            assert oaio_id in mocked_client.client_objectz
            assert mocked_client.client_objectz[oaio_id] == oai_obj

            assert mocked_client.o_synchronize_with_server_if_online()
            for fil_nam in file_paths:
                assert csh_api.deployed_file_content(os_path_join(svr_path, fil_nam)) == fil_content + fil_nam.encode()

        finally:
            if os_path_isdir(root_path):
                shutil.rmtree(root_path)
            err_msg = "ERR:" + mocked_client.error_message
            if oaio_id:
                assert mocked_client.unregister_object(oaio_id, wipe_files=True) == "", err_msg
                assert len(mocked_client.client_objectz) == local_object_count, err_msg

    def test_register_object(self, mocked_client):
        oai_obj = None
        try:
            oai_obj = mocked_client.register_object({})
            assert mocked_client.o_synchronize_with_server_if_online()

            assert not mocked_client.error_message
            assert oai_obj
            assert oai_obj.oaio_id
            assert oai_obj.client_stamp
            assert oai_obj.client_values == {}
            assert oai_obj.csh_id == CSH_API_ID

            assert not deployed_file_contents, "ERR:" + mocked_client.error_message

            path = os_path_join(mocked_client.client_root, OBJECTS_DIR, oai_obj.oaio_id)
            assert os_path_isfile(path)

            oai_obj = mocked_client._load_client_object_info(oai_obj.oaio_id)
            assert oai_obj.oaio_id == oai_obj.oaio_id
            assert oai_obj.client_stamp == oai_obj.client_stamp
            assert oai_obj.client_values == oai_obj.client_values
            assert oai_obj.csh_id == oai_obj.csh_id

        finally:
            if oai_obj:
                mocked_client.unregister_object(oai_obj.oaio_id, wipe_files=True)
                assert not os_path_isfile(os_path_join(mocked_client.client_root, OBJECTS_DIR, oai_obj.oaio_id))

    def test_synchronize_with_server_if_online(self, mocked_client):
        assert mocked_client.o_synchronize_with_server_if_online() is True

    def test_synchronize_with_server_if_online_with_2nd_usr(self, mocked_client, other_client):
        reg_stamp = now_stamp()
        reg_val = {'reg_key': 'reg_val'}

        oai_obj = mocked_client.register_object(reg_val, stamp=reg_stamp)
        assert oai_obj and mocked_client.o_synchronize_with_server_if_online()
        oaio_id = oai_obj.oaio_id

        syn_obj = mocked_client.client_objectz[oaio_id]
        assert syn_obj.client_stamp == reg_stamp
        assert syn_obj.server_stamp == reg_stamp
        assert syn_obj.client_values == reg_val
        assert syn_obj.server_values == reg_val
        assert syn_obj == mocked_client._load_client_object_info(oaio_id)

        assert mocked_client.upsert_subscription(oaio_id, other_client.user_name)
        assert other_client.o_synchronize_with_server_if_online()
        upd_stamp = now_stamp()
        upd_val = {'upd_key': 'upd_val'}

        _upd_obj = other_client.update_object(oaio_id, upd_val, stamp=upd_stamp, reset=True)

        syn_obj = other_client.client_objectz[oaio_id]
        assert syn_obj.client_stamp == upd_stamp
        assert syn_obj.server_stamp == reg_stamp
        assert syn_obj.client_values == upd_val
        assert syn_obj.server_values == reg_val
        assert syn_obj == other_client._load_client_object_info(oaio_id)

        assert other_client.o_synchronize_with_server_if_online()   # 2nd user uploading updated values to server

        syn_obj = other_client.client_objectz[oaio_id]
        assert syn_obj.client_stamp == upd_stamp
        assert syn_obj.server_stamp == upd_stamp
        assert syn_obj.client_values == upd_val
        assert syn_obj.server_values == upd_val
        assert syn_obj == other_client._load_client_object_info(oaio_id)

        assert mocked_client.o_synchronize_with_server_if_online()  # registering creator updated by other_client

        syn_obj = mocked_client.client_objectz[oaio_id]
        assert syn_obj.client_stamp == upd_stamp
        assert syn_obj.server_stamp == upd_stamp
        assert syn_obj.client_values == upd_val
        assert syn_obj.server_values == upd_val
        assert syn_obj == mocked_client._load_client_object_info(oaio_id)

    def test_synchronize_with_server_if_online_with_2nd_usr_file(self, mocked_client, other_client):
        reg_stamp = now_stamp()
        reg_name = 'reg_syn_with_download.tst'
        reg_content = b"register content of tst_syn_with_file"

        oai_obj = mocked_client.register_file(reg_name, file_content=reg_content, stamp=reg_stamp)
        assert oai_obj and mocked_client.o_synchronize_with_server_if_online()
        oaio_id = oai_obj.oaio_id
        reg_val = oai_obj.client_values
        fil_root = os_path_join(mocked_client.client_root, FILES_DIR, oaio_id)
        oth_root = os_path_join(other_client.client_root, FILES_DIR, oaio_id)

        syn_obj = mocked_client.client_objectz[oaio_id]
        assert syn_obj.client_stamp == reg_stamp
        assert syn_obj.server_stamp == reg_stamp
        assert syn_obj.client_values == reg_val
        assert syn_obj.server_values == reg_val
        assert syn_obj == mocked_client._load_client_object_info(oaio_id)
        # noinspection PyTypeChecker
        assert read_file(os_path_join(fil_root, reg_name), extra_mode='b') == reg_content

        assert mocked_client.upsert_subscription(oaio_id, other_client.user_name)
        assert other_client.o_synchronize_with_server_if_online()
        upd_stamp = now_stamp()
        upd_name = 'upd_syn_file.name'
        upd_content = b"updated synFileContent"

        upd_obj = other_client.update_file(oaio_id, file_name=upd_name, file_content=upd_content, stamp=upd_stamp)

        assert upd_obj
        upd_val = upd_obj.client_values
        syn_obj = other_client.client_objectz[oaio_id]
        assert syn_obj.oaio_id == oaio_id
        assert syn_obj.client_stamp == upd_stamp
        assert syn_obj.server_stamp == reg_stamp
        assert syn_obj.client_values == upd_val
        assert syn_obj.server_values == reg_val
        assert syn_obj == other_client._load_client_object_info(oaio_id)
        # noinspection PyTypeChecker
        assert read_file(os_path_join(oth_root, upd_name), extra_mode='b') == upd_content

        assert other_client.o_synchronize_with_server_if_online()

        syn_obj = other_client.client_objectz[oaio_id]
        assert syn_obj.client_stamp == upd_stamp
        assert syn_obj.server_stamp == upd_stamp
        assert syn_obj.client_values == upd_val
        assert syn_obj.server_values == upd_val
        assert syn_obj == other_client._load_client_object_info(oaio_id)
        # noinspection PyTypeChecker
        assert read_file(os_path_join(oth_root, upd_name), extra_mode='b') == upd_content

        assert mocked_client.o_synchronize_with_server_if_online()

        syn_obj = mocked_client.client_objectz[oaio_id]
        assert syn_obj.client_stamp == upd_stamp
        assert syn_obj.server_stamp == upd_stamp
        assert syn_obj.client_values == upd_val
        assert syn_obj.server_values == upd_val
        assert syn_obj == mocked_client._load_client_object_info(oaio_id)
        # noinspection PyTypeChecker
        assert read_file(os_path_join(fil_root, upd_name), extra_mode='b') == upd_content

    def test_unregister_object(self, mocked_client):
        oai_obj = mocked_client.register_object({})
        assert oai_obj
        assert oai_obj.oaio_id in mocked_client.client_objectz
        assert mocked_client.o_synchronize_with_server_if_online()

        assert mocked_client.unregister_object(oai_obj.oaio_id) == ""

        assert oai_obj.oaio_id not in mocked_client.client_objectz
        assert not os_path_isfile(os_path_join(mocked_client.client_root, OBJECTS_DIR, oai_obj.oaio_id))

    def test_unregister_object_not_registered_on_servers(self, mocked_client):
        oai_obj = mocked_client.register_object({})
        assert oai_obj
        assert oai_obj.oaio_id in mocked_client.client_objectz

        assert mocked_client.unregister_object(oai_obj.oaio_id) == ""

        assert oai_obj.oaio_id not in mocked_client.client_objectz
        assert not os_path_isfile(os_path_join(mocked_client.client_root, OBJECTS_DIR, oai_obj.oaio_id))

    def test_unregister_object_file_wipe(self, mocked_client):
        reg_name = 'tst_unregister_wiped_file.txt'
        oai_obj = mocked_client.register_file(reg_name, file_content=b"unregister wipe err test file content")
        assert mocked_client.o_synchronize_with_server_if_online()

        err_msg = mocked_client.unregister_object(oai_obj.oaio_id, wipe_files=True)

        assert not err_msg
        assert oai_obj.oaio_id not in mocked_client.client_objectz
        assert not os_path_isfile(os_path_join(mocked_client.client_root, OBJECTS_DIR, oai_obj.oaio_id))
        assert not os_path_isdir(os_path_join(mocked_client.client_root, FILES_DIR, oai_obj.oaio_id))
        assert not os_path_isfile(os_path_join(mocked_client.client_root, FILES_DIR, oai_obj.oaio_id, reg_name))
        assert _csh_file_content(mocked_client, oai_obj) == {reg_name: None}

    def test_unregister_object_file_wipe_err(self, mocked_client):
        reg_name = 'tst_unregister_wiped_file.txt'
        oai_obj = mocked_client.register_file(reg_name, file_content=b"unregister wipe err test file content")
        assert mocked_client.o_synchronize_with_server_if_online()
        assert _csh_file_remove(mocked_client, oai_obj) == ""  # force error by removing file from the cloud storage
        assert _csh_mock_cleared_errors() == ""

        err_msg = mocked_client.unregister_object(oai_obj.oaio_id, wipe_files=True)

        assert reg_name in err_msg
        assert oai_obj.oaio_id in err_msg

    def test_unregister_object_map_err(self, mocked_client):
        oai_obj = mocked_client.register_object({})
        mocked_client.client_objectz.pop(oai_obj.oaio_id)

        err_msg = mocked_client.unregister_object(oai_obj.oaio_id)

        assert oai_obj.oaio_id in err_msg

    def test_unregister_server_err(self, mocked_client):
        oai_obj = mocked_client.register_object({})
        assert mocked_client.o_synchronize_with_server_if_online()
        mocked_client.session.post.side_effect = _mocked_get_or_post_failure_response

        err_msg = mocked_client.unregister_object(oai_obj.oaio_id)

        assert oai_obj.oaio_id in err_msg

    def test_unsubscribe(self, mocked_client):
        oai_obj = mocked_client.register_object({})

        assert mocked_client.unsubscribe(oai_obj.oaio_id, oai_obj.username) == ""

        assert oai_obj.oaio_id not in mocked_client.client_objectz
        assert not os_path_isfile(os_path_join(mocked_client.client_root, OBJECTS_DIR, oai_obj.oaio_id))

    def test_unsubscribe_client_obj_file_err(self, mocked_client):
        oai_obj = mocked_client.register_object({})
        os.remove(os_path_join(mocked_client.client_root, OBJECTS_DIR, oai_obj.oaio_id))

        err_msg = mocked_client.unsubscribe(oai_obj.oaio_id, oai_obj.username)

        assert oai_obj.oaio_id in err_msg
        assert oai_obj.username in err_msg

    def test_unsubscribe_client_obj_map_err(self, mocked_client):
        oai_obj = mocked_client.register_object({})
        mocked_client.client_objectz.pop(oai_obj.oaio_id)

        err_msg = mocked_client.unsubscribe(oai_obj.oaio_id, oai_obj.username)

        assert oai_obj.oaio_id in err_msg
        assert oai_obj.username in err_msg

    def test_unsubscribe_server_err(self, mocked_client):
        oai_obj = mocked_client.register_object({})
        mocked_client.session.post.side_effect = _mocked_get_or_post_failure_response

        err_msg = mocked_client.unsubscribe(oai_obj.oaio_id, oai_obj.username)

        assert oai_obj.oaio_id in err_msg
        assert oai_obj.username in err_msg

    def test_update_file(self, mocked_client):
        oai_obj = None
        dir_path = os_path_abspath(os_path_join(TESTS_FOLDER, 'tst_file_dir'))
        fil_name = 'tst_fil.tst'
        fil_path = os_path_join(dir_path, fil_name)
        fil_content = b"update test file \x00 having binary content \x01\x02\x03\xff"

        try:
            reg_stamp = now_stamp()
            write_file(fil_path, b"", make_dirs=True)
            oai_obj = mocked_client.register_file(fil_name, root_path=dir_path, stamp=reg_stamp)
            assert mocked_client.error_message == ""
            assert oai_obj is not None
            assert mocked_client.o_synchronize_with_server_if_online()
            assert _csh_file_content(mocked_client, oai_obj) == {fil_name: b""}

            write_file(fil_path, fil_content)
            upd_stamp = now_stamp()

            oai_obj = mocked_client.update_file(oai_obj.oaio_id, stamp=upd_stamp)
            assert mocked_client.o_synchronize_with_server_if_online()

            assert ROOT_VALUES_KEY in oai_obj.client_values
            assert normalize(oai_obj.client_values[ROOT_VALUES_KEY]) == dir_path
            assert FILES_VALUES_KEY in oai_obj.client_values
            assert len(oai_obj.client_values[FILES_VALUES_KEY]) == 1
            assert oai_obj.client_values[FILES_VALUES_KEY][0] == fil_name
            assert oai_obj.client_stamp == upd_stamp
            assert oai_obj.server_stamp == reg_stamp
            assert oai_obj.client_values == oai_obj.server_values
            assert oai_obj.csh_access_right == CREATE_ACCESS_RIGHT

            assert len(mocked_client.client_objectz) == 1
            assert oai_obj.oaio_id in mocked_client.client_objectz

            assert os_path_isfile(fil_path)
            assert _csh_file_content(mocked_client, oai_obj) == {fil_name: fil_content}

        finally:
            if os_path_isdir(dir_path):
                shutil.rmtree(dir_path)
            if oai_obj:
                assert mocked_client.unregister_object(oai_obj.oaio_id) == "", "ERR:" + mocked_client.error_message
                assert len(mocked_client.client_objectz) == 0

    def test_update_file_and_root(self, mocked_client):
        oai_obj = None
        reg_root = mocked_client.client_root
        upd_root = os_path_join(TESTS_FOLDER, 'upd_root')
        reg_name = 'reg_fil.tst'
        upd_name = 'upd_fil.tst'
        reg_path = os_path_join(reg_root, reg_name)
        upd_path = os_path_join(upd_root, upd_name)
        fil_content = b"update test file \x00 having binary content \x01\x02\x03\xff"
        reg_stamp = now_stamp()

        try:
            write_file(reg_path, b"")
            oai_obj = mocked_client.register_file(reg_name, root_path=reg_root, stamp=reg_stamp)
            assert mocked_client.error_message == ""
            assert oai_obj is not None
            assert mocked_client.o_synchronize_with_server_if_online()
            assert _csh_file_content(mocked_client, oai_obj) == {reg_name: b""}
            oaio_id = oai_obj.oaio_id

            write_file(upd_path, fil_content, make_dirs=True)
            upd_stamp = now_stamp()

            oai_obj = mocked_client.update_file(oaio_id, file_name=upd_name, root_path=upd_root, stamp=upd_stamp)
            assert mocked_client.o_synchronize_with_server_if_online()

            assert ROOT_VALUES_KEY in oai_obj.client_values
            assert oai_obj.client_values[ROOT_VALUES_KEY] == upd_root
            assert oai_obj.server_values[ROOT_VALUES_KEY] == placeholder_path(reg_root)
            assert FILES_VALUES_KEY in oai_obj.client_values
            assert len(oai_obj.client_values[FILES_VALUES_KEY]) == 1
            assert oai_obj.client_values[FILES_VALUES_KEY][0] == upd_name
            assert oai_obj.server_values[FILES_VALUES_KEY][0] == reg_name
            assert oai_obj.client_stamp == upd_stamp
            assert oai_obj.server_stamp == reg_stamp
            assert oai_obj.csh_access_right == CREATE_ACCESS_RIGHT

            assert len(mocked_client.client_objectz) == 1
            assert oaio_id in mocked_client.client_objectz

            assert os_path_isfile(reg_path)
            assert os_path_isfile(upd_path)
            assert _csh_file_content(mocked_client, oai_obj) == {upd_name: fil_content}

        finally:
            if oai_obj:
                assert mocked_client.unregister_object(oai_obj.oaio_id) == "", "ERR:" + mocked_client.error_message
                assert _csh_file_remove(mocked_client, oai_obj) == ""
                assert _csh_mock_cleared_errors() == ""
                assert len(mocked_client.client_objectz) == 0
            if os_path_isdir(reg_root):
                shutil.rmtree(reg_root)
            if os_path_isdir(upd_root):
                shutil.rmtree(upd_root)

    def test_update_folder(self, mocked_client):
        # noinspection PyTypeChecker
        root_path = os_path_abspath(os_path_join(mocked_client.client_root, FILES_DIR))
        sub_dirs = ['sub_dir1', 'sub_dir3/sub sub_dir']
        file_paths = ['tEst file.tst',
                      os_path_join(sub_dirs[0], 'tst sub1-fil.tst'),
                      os_path_join(sub_dirs[1], 'tst_sub3_sub_fil.tst'),
                      ]
        fil_content = b"update folder test file \x00 binary content \x01\x02\x03\xff"

        assert mocked_client.o_synchronize_with_server_if_online()
        local_object_count = len(mocked_client.client_objectz)
        csh_api = mocked_client._csh_api(mocked_client.csh_default_id)
        oai_obj = None
        oaio_id = ""

        try:
            reg_stamp = now_stamp()
            for dir_nam in sub_dirs:
                os.makedirs(os_path_join(root_path, dir_nam))
            oai_obj = mocked_client.register_folder(root_path, stamp=reg_stamp)
            assert mocked_client.o_synchronize_with_server_if_online()
            assert mocked_client.error_message == ""
            assert len(mocked_client.client_objectz[oai_obj.oaio_id].client_values[FILES_VALUES_KEY]) == 0
            oaio_id = oai_obj.oaio_id

            upd_stamp = now_stamp()
            for fil_nam in file_paths:
                write_file(os_path_join(root_path, fil_nam), fil_content + fil_nam.encode())

            oai_obj = mocked_client.update_folder(oai_obj.oaio_id, root_path=root_path, stamp=upd_stamp)

            assert mocked_client.error_message == ""
            assert oai_obj.oaio_id in mocked_client.client_objectz
            assert oai_obj is mocked_client.client_objectz[oai_obj.oaio_id]
            assert oai_obj.client_stamp == upd_stamp
            assert len(oai_obj.client_values[FILES_VALUES_KEY]) == len(file_paths)
            for fil_nam in file_paths:
                assert fil_nam in oai_obj.client_values[FILES_VALUES_KEY]
                assert os_path_isfile(normalize(os_path_join(root_path, fil_nam)))
                srv_fil = os_path_join(oaio_id, upd_stamp, fil_nam)
                assert csh_api.deployed_file_content(srv_fil) is None
            assert oai_obj.server_stamp == reg_stamp
            assert oai_obj.server_values.get(FILES_VALUES_KEY) == []

            assert mocked_client.o_synchronize_with_server_if_online()

            assert mocked_client.error_message == ""
            assert oai_obj.oaio_id in mocked_client.client_objectz
            assert oai_obj.client_stamp == upd_stamp
            assert len(oai_obj.client_values[FILES_VALUES_KEY]) == len(file_paths)
            for fil_nam in file_paths:
                assert fil_nam in oai_obj.client_values[FILES_VALUES_KEY]
                assert os_path_isfile(normalize(os_path_join(root_path, fil_nam)))
                srv_fil = os_path_join(oaio_id, upd_stamp, fil_nam)
                assert csh_api.deployed_file_content(srv_fil) == fil_content + fil_nam.encode()

            assert oai_obj.server_stamp == reg_stamp
            assert oai_obj.server_values.get(FILES_VALUES_KEY) == []

        finally:
            err_msg = "ERR:" + mocked_client.error_message
            if oaio_id:
                assert mocked_client.unregister_object(oaio_id, wipe_files=True) == "", err_msg
                assert not os_path_isfile(os_path_join(mocked_client.client_root, OBJECTS_DIR, oai_obj.oaio_id))
                assert len(mocked_client.client_objectz) == local_object_count, err_msg

    def test_update_object_after_upload(self, mocked_client):
        oai_obj = None
        try:
            reg_stamp = now_stamp()
            oai_obj = mocked_client.register_object({'val_key': 'reg_val'}, stamp=reg_stamp)
            assert mocked_client.o_synchronize_with_server_if_online()
            upd_stamp = now_stamp()

            mocked_client.update_object(oai_obj.oaio_id, {'val_key': 'upd_val'}, stamp=upd_stamp)

            assert oai_obj.oaio_id in mocked_client.client_objectz
            assert mocked_client.client_objectz[oai_obj.oaio_id].client_stamp == upd_stamp
            assert mocked_client.client_objectz[oai_obj.oaio_id].client_values['val_key'] == 'upd_val'

            assert mocked_client.client_objectz[oai_obj.oaio_id].server_stamp == reg_stamp
            assert mocked_client.client_objectz[oai_obj.oaio_id].server_values.get('val_key') == 'reg_val'

        finally:
            if oai_obj:
                assert mocked_client.unregister_object(oai_obj.oaio_id, wipe_files=True) == ""
                assert not os_path_isfile(os_path_join(mocked_client.client_root, OBJECTS_DIR, oai_obj.oaio_id))

    def test_update_object_before_upload(self, mocked_client):
        oai_obj = None
        try:
            reg_stamp = now_stamp()
            oai_obj = mocked_client.register_object({'val_key': 'reg_val'}, stamp=reg_stamp)
            upd_stamp = now_stamp()

            mocked_client.update_object(oai_obj.oaio_id, {'val_key': 'upd_val'}, stamp=upd_stamp)

            assert oai_obj.oaio_id in mocked_client.client_objectz
            assert mocked_client.client_objectz[oai_obj.oaio_id].client_stamp == upd_stamp
            assert mocked_client.client_objectz[oai_obj.oaio_id].client_values['val_key'] == 'upd_val'

            assert not mocked_client.client_objectz[oai_obj.oaio_id].server_stamp
            assert not mocked_client.client_objectz[oai_obj.oaio_id].server_values

        finally:
            if oai_obj:
                assert mocked_client.unregister_object(oai_obj.oaio_id, wipe_files=True) == ""
                assert not os_path_isfile(os_path_join(mocked_client.client_root, OBJECTS_DIR, oai_obj.oaio_id))

    def test_upsert_subscription(self, mocked_client):
        oai_obj = mocked_client.register_object({})

        assert oai_obj
        assert oai_obj.oaio_id in mocked_client.client_objectz
        assert oai_obj.username == mocked_client.user_name
        assert oai_obj.csh_access_right == CREATE_ACCESS_RIGHT

        pid = 0
        try:
            pid = mocked_client.upsert_subscription(oai_obj.oaio_id, other_user, access_right=UPDATE_ACCESS_RIGHT)
            assert mocked_client.error_message == ""
            assert pid != 0

            mocked_client.user_name = mocked_client.session.headers[HTTP_HEADER_USR_ID] = other_user
            assert oai_obj.oaio_id in mocked_client.client_objectz

        finally:
            if pid:
                mocked_client.user_name = mocked_client.session.headers[HTTP_HEADER_USR_ID] = oai_obj.username
                assert mocked_client.unsubscribe(oai_obj.oaio_id, other_user) == ""
                assert oai_obj.oaio_id in mocked_client.client_objectz

                mocked_client.user_name = mocked_client.session.headers[HTTP_HEADER_USR_ID] = other_user

    def test_upsert_subscription_err(self, mocked_client):
        oai_obj = mocked_client.register_object({})
        mocked_client.session.post.side_effect = _mocked_get_or_post_failure_response

        pid = mocked_client.upsert_subscription(oai_obj.oaio_id, other_user, access_right=UPDATE_ACCESS_RIGHT)

        assert pid == 0
        assert oai_obj.oaio_id in mocked_client.error_message
        assert other_user in mocked_client.error_message

    def test_userz_access(self, mocked_client):
        users = mocked_client.userz_access('any_id_because_only_checking_for_correct_user_name')
        assert isinstance(users, list)
        assert len(users) == 1
        assert users[0]['username'] == 'mocked_username'

    def test_userz_access_err(self, mocked_client):
        mocked_client.session.post.side_effect = _mocked_get_or_post_failure_response
        users = mocked_client.userz_access('any_id_because_only_checking_for_correct_user_name')
        assert isinstance(users, list)
        assert len(users) == 0

    def test_wipe_user(self, mocked_client):
        oai_obj = None
        try:
            oai_obj = mocked_client.register_object({})
            mocked_client.upsert_subscription(oai_obj.oaio_id, other_user)  # create User/Userz recs for other usr

            assert mocked_client.wipe_user(other_user) == ""

        finally:
            if oai_obj:
                assert mocked_client.unregister_object(oai_obj.oaio_id) == ""


# ******   integration test against live cloud storage and oaio servers   ******


@pytest.fixture(scope="session")
def live_client():
    """ on local machine connect to personal server in order to do integration tests """
    client_root = os_path_join(TESTS_FOLDER, 'tst_liv_rud')
    assert not os_path_isdir(client_root)

    ocl = OaioClient(os.environ['OAIO_HOST_NAME'],
                     {'username': os.environ['OAIO_USERNAME'], 'password': os.environ['OAIO_PASSWORD']},
                     app_id='aIi',
                     device_id='dIi',
                     csh_default_id='_TstCshLive',      # actually not needed because _csh_api() get monkey patched
                     client_root=client_root,
                     )
    ocl.o_synchronize_with_server_if_online = ocl.synchronize_with_server_if_online
    ocl.synchronize_with_server_if_online = lambda: True  # monkey patch to allow manual server sync in tests

    liv_api = DigiApi(root_folder='tst_liv_rut',        # TEST_DIGI_ROOT_FOLDER_PATH exclusive to test_cloud_storage.py
                      email=os.environ.get('TEST_DIGI_API_EMAIL'),
                      password=os.environ.get('TEST_DIGI_API_PASSWORD'))
    ocl._csh_api = lambda *_args: liv_api               # monkey patch to prevent Cshz test rec in live server database

    assert ocl.error_message == ""
    local_obj_cache_path = os_path_join(ocl.client_root, OBJECTS_DIR)

    # noinspection PyProtectedMember,PyTypeChecker
    assert not os.listdir(local_obj_cache_path) and not ocl._load_server_objectz()  # ensure empty/clean test env
    assert other_user not in [_['username'] for _ in ocl.userz_access('any_existing_or_not_existing_oaio_id')]

    yield ocl

    # noinspection PyProtectedMember,PyTypeChecker
    assert not os.listdir(local_obj_cache_path) and not ocl._load_server_objectz()  # ensure empty/clean test env
    all_users = [_['username'] for _ in ocl.userz_access('any_existing_or_not_existing_oaio_id')]
    assert other_user not in all_users
    shutil.rmtree(client_root)


@pytest.mark.skipif("not os.environ.get('OAIO_HOST_NAME')", reason="disable tests on gitlab CI or if servers are down")
class TestIntegrationWithLiveServers:
    """ integration tests running only on local machine, using your private oaio and cloud storage live servers. """
    def test_connect(self, live_client):
        assert live_client.error_message == ""
        assert live_client.connected

        assert live_client.o_synchronize_with_server_if_online(), f"{live_client.error_message=}"
        assert live_client.connected

    def test_last_sync_stamp(self, live_client):
        assert live_client.last_sync_stamp
        stamp = now_stamp()
        live_client.last_sync_stamp = stamp
        assert live_client.last_sync_stamp == stamp

    def test_register_file(self, live_client):
        oai_obj = None
        stamp = now_stamp()
        dir_path = ""
        fil_name = 'tst_fil.yyy'
        fil_content = b"test file \x00 binary content \x01\x02\x03\xff"

        assert live_client.o_synchronize_with_server_if_online(), f"{live_client.error_message=}"
        local_object_count = len(live_client.client_objectz)

        try:
            oai_obj = live_client.register_file(fil_name, file_content=fil_content, stamp=stamp)

            oaio_id = oai_obj.oaio_id
            dir_path = live_client._client_file_root(oaio_id, oai_obj.client_values)
            assert live_client.error_message == ""
            assert oai_obj is not None
            assert oaio_id
            assert oai_obj.csh_id

            assert ROOT_VALUES_KEY not in oai_obj.client_values
            assert FILES_VALUES_KEY in oai_obj.client_values
            assert len(oai_obj.client_values[FILES_VALUES_KEY]) == 1
            assert oai_obj.client_values[FILES_VALUES_KEY][0] == fil_name
            assert os_path_isfile(os_path_join(dir_path, fil_name))
            assert not oai_obj.server_values
            assert oai_obj.client_stamp == stamp
            assert not oai_obj.server_stamp
            assert oai_obj.csh_access_right == CREATE_ACCESS_RIGHT

            assert len(live_client.client_objectz) == local_object_count + 1
            assert oaio_id in live_client.client_objectz
            assert live_client.client_objectz[oaio_id] == oai_obj

            assert _csh_file_content(live_client, oai_obj) == {fil_name: None}

            assert live_client.o_synchronize_with_server_if_online(), f"{live_client.error_message=}"
            assert _csh_file_content(live_client, oai_obj) == {fil_name: fil_content}

        finally:
            if oai_obj:
                assert _csh_file_remove(live_client, oai_obj) == ""
                assert live_client.unregister_object(oai_obj.oaio_id) == ""
                assert len(live_client.client_objectz) == local_object_count
            if os_path_isdir(dir_path):
                shutil.rmtree(dir_path)

    def test_register_file_with_root(self, live_client):
        oai_obj = None
        stamp = now_stamp()
        dir_path = os_path_abspath(os_path_join(TESTS_FOLDER, 'reg file dir'))
        fil_name = 'tst_fil.tst'
        fil_path = os_path_join(dir_path, fil_name)
        fil_content = b"test file \x00 having binary content \x01\x02\x03\xff"

        local_object_count = len(live_client.client_objectz)

        try:
            write_file(fil_path, fil_content, make_dirs=True)

            oai_obj = live_client.register_file(fil_name, root_path=dir_path, stamp=stamp)
            assert live_client.error_message == ""
            assert oai_obj is not None

            assert oai_obj.oaio_id
            assert oai_obj.csh_id
            assert ROOT_VALUES_KEY in oai_obj.client_values
            assert normalize(oai_obj.client_values[ROOT_VALUES_KEY]) == dir_path
            assert FILES_VALUES_KEY in oai_obj.client_values
            assert len(oai_obj.client_values[FILES_VALUES_KEY]) == 1
            assert oai_obj.client_values[FILES_VALUES_KEY][0] == fil_name
            assert not oai_obj.server_values
            assert oai_obj.client_stamp == stamp
            assert not oai_obj.server_stamp
            assert oai_obj.csh_access_right == CREATE_ACCESS_RIGHT

            assert len(live_client.client_objectz) == local_object_count + 1
            assert oai_obj.oaio_id in live_client.client_objectz
            assert live_client.client_objectz[oai_obj.oaio_id] == oai_obj

            assert os_path_isfile(fil_path)
            assert _csh_file_content(live_client, oai_obj) == {fil_name: None}

            assert live_client.o_synchronize_with_server_if_online(), f"{live_client.error_message=}"
            assert _csh_file_content(live_client, oai_obj) == {fil_name: fil_content}

        finally:
            if oai_obj:
                assert _csh_file_remove(live_client, oai_obj) == ""
                assert live_client.unregister_object(oai_obj.oaio_id) == ""
                assert len(live_client.client_objectz) == local_object_count
            if os_path_isdir(dir_path):
                shutil.rmtree(dir_path)

    def test_register_folder_with_root(self, live_client):
        stamp = now_stamp()
        folder_path = os_path_abspath(os_path_join(TESTS_FOLDER, 'folder_root'))
        sub_dirs = ['sub_dir1', 'sub_dir3/sub_sub_dir']
        file_paths = ['tst_fil.tst',
                      os_path_join(sub_dirs[0], 'tst_sub1_fil.tst'),
                      os_path_join(sub_dirs[1], 'tst_sub3_sub_fil.tst'),
                      ]
        fil_content = b"test file \x00 binary content \x01\x02\x03\xff"

        local_object_count = len(live_client.client_objectz)
        csh_api = live_client._csh_api(live_client.csh_default_id)
        oai_obj = None
        oaio_id = ""

        try:
            for fil_nam in file_paths:
                write_file(os_path_join(folder_path, fil_nam), fil_content + fil_nam.encode(), make_dirs=True)

            oai_obj = live_client.register_folder(folder_path, stamp=stamp)
            oaio_id = oai_obj.oaio_id

            assert oai_obj is not None
            assert oai_obj.oaio_id
            assert live_client.error_message == ""
            assert oai_obj.csh_id
            assert ROOT_VALUES_KEY in oai_obj.client_values
            assert normalize(oai_obj.client_values[ROOT_VALUES_KEY]) == normalize(folder_path)
            assert FILES_VALUES_KEY in oai_obj.client_values
            assert len(oai_obj.client_values[FILES_VALUES_KEY]) == len(file_paths)
            svr_path = os_path_join(oaio_id, stamp)
            for fil_nam in file_paths:
                assert fil_nam in oai_obj.client_values[FILES_VALUES_KEY]
                assert os_path_isfile(normalize(os_path_join(folder_path, fil_nam)))
                assert csh_api.deployed_file_content(os_path_join(svr_path, fil_nam)) is None
            assert oai_obj.client_stamp == stamp
            assert not oai_obj.server_stamp
            assert not oai_obj.server_values
            assert oai_obj.csh_access_right == CREATE_ACCESS_RIGHT
            assert len(live_client.client_objectz) == local_object_count + 1
            assert oaio_id in live_client.client_objectz
            assert live_client.client_objectz[oaio_id] == oai_obj

            assert live_client.o_synchronize_with_server_if_online(), f"{live_client.error_message=}"

            assert live_client.error_message == ""
            oai_obj = live_client.client_objectz.get(oai_obj.oaio_id)
            assert oai_obj
            assert oai_obj.oaio_id
            assert oai_obj.csh_id
            assert ROOT_VALUES_KEY in oai_obj.client_values
            assert normalize(oai_obj.client_values[ROOT_VALUES_KEY]) == normalize(folder_path)
            assert FILES_VALUES_KEY in oai_obj.client_values
            assert len(oai_obj.client_values[FILES_VALUES_KEY]) == len(file_paths)
            svr_path = os_path_join(oaio_id, stamp)
            for fil_nam in file_paths:
                assert fil_nam in oai_obj.client_values[FILES_VALUES_KEY]
                assert os_path_isfile(normalize(os_path_join(folder_path, fil_nam)))
                assert csh_api.deployed_file_content(os_path_join(svr_path, fil_nam)) == fil_content + fil_nam.encode()
            assert oai_obj.client_stamp == stamp
            assert oai_obj.server_stamp == stamp
            assert oai_obj.server_values == oai_obj.client_values
            assert oai_obj.csh_access_right == CREATE_ACCESS_RIGHT
            assert len(live_client.client_objectz) == local_object_count + 1

        finally:
            if oaio_id:
                assert live_client.unregister_object(oaio_id, wipe_files=True) == ""
                assert len(live_client.client_objectz) == local_object_count
                err_msg = _csh_file_remove(live_client, oai_obj, file_count=len(file_paths))
                assert stamp in err_msg
                assert oaio_id in err_msg
            if os_path_isdir(folder_path):
                shutil.rmtree(folder_path)

    def test_register_object(self, live_client):
        oai_obj = None
        stamp = now_stamp()
        values = {'tst_str': 'tst_val', 'tst_int': 69}

        local_object_count = len(live_client.client_objectz)

        try:
            oai_obj = live_client.register_object(values, stamp=stamp)
            assert live_client.error_message == ""
            assert oai_obj is not None

            assert oai_obj.oaio_id
            assert oai_obj.csh_id
            assert oai_obj.client_values == values
            assert not oai_obj.server_values
            assert oai_obj.client_stamp == stamp
            assert not oai_obj.server_stamp
            assert oai_obj.csh_access_right == CREATE_ACCESS_RIGHT
            assert len(live_client.client_objectz) == local_object_count + 1
            assert oai_obj.oaio_id in live_client.client_objectz
            assert live_client.client_objectz[oai_obj.oaio_id] == oai_obj

            assert live_client.o_synchronize_with_server_if_online(), f"{live_client.error_message=}"

            assert live_client.error_message == ""
            oai_obj = live_client.client_objectz.get(oai_obj.oaio_id)
            assert oai_obj
            assert oai_obj.oaio_id
            assert oai_obj.csh_id
            assert oai_obj.client_values == values
            assert oai_obj.server_values == values
            assert oai_obj.client_stamp == stamp
            assert oai_obj.server_stamp == stamp
            assert oai_obj.csh_access_right == CREATE_ACCESS_RIGHT
            assert len(live_client.client_objectz) == local_object_count + 1

        finally:
            if oai_obj:
                assert live_client.unregister_object(oai_obj.oaio_id) == ""
                assert len(live_client.client_objectz) == local_object_count

    def test_synchronize_with_server_if_online(self, live_client):
        assert live_client.o_synchronize_with_server_if_online() is True, f"{live_client.error_message=}"

    def test_unregister_object(self, live_client):
        oai_obj = live_client.register_object({})
        assert live_client.o_synchronize_with_server_if_online(), f"{live_client.error_message=}"

        assert oai_obj
        assert oai_obj.oaio_id in live_client._load_server_objectz()
        assert oai_obj.oaio_id in live_client.client_objectz
        assert oai_obj.username == live_client.user_name
        assert oai_obj.csh_access_right == CREATE_ACCESS_RIGHT

        pid = 0
        try:
            pid = live_client.upsert_subscription(oai_obj.oaio_id, other_user, access_right=UPDATE_ACCESS_RIGHT)
            assert live_client.error_message == ""
            assert pid != 0

            # to allow check the subscription of other user and to mock a request, OAIO_USERNAME has to be supervisor
            live_client.user_name = live_client.session.headers[HTTP_HEADER_USR_ID] = other_user
            assert oai_obj.oaio_id in live_client._load_server_objectz(), f"{live_client.error_message=}"

        finally:
            if pid:
                live_client.user_name = live_client.session.headers[HTTP_HEADER_USR_ID] = oai_obj.username
                assert live_client.unsubscribe(oai_obj.oaio_id, other_user) == ""
                assert oai_obj.oaio_id in live_client._load_server_objectz(), f"{live_client.error_message=}"

                live_client.user_name = live_client.session.headers[HTTP_HEADER_USR_ID] = other_user
                assert oai_obj.oaio_id not in live_client._load_server_objectz()

                live_client.user_name = live_client.session.headers[HTTP_HEADER_USR_ID] = oai_obj.username
                assert live_client.unsubscribe(oai_obj.oaio_id, oai_obj.username)   # creator subscription not deletable
                assert live_client.unregister_object(oai_obj.oaio_id) == ""
                assert oai_obj.oaio_id not in live_client._load_server_objectz(), f"{live_client.error_message=}"

    def test_unsubscribe(self, live_client):
        oai_obj = None
        try:
            oai_obj = live_client.register_object({})
            assert live_client.o_synchronize_with_server_if_online(), f"{live_client.error_message=}"
            assert live_client.upsert_subscription(oai_obj.oaio_id, other_user)

            assert oai_obj
            assert oai_obj.oaio_id in live_client._load_server_objectz()
            assert oai_obj.oaio_id in live_client.client_objectz

            assert live_client.unsubscribe(oai_obj.oaio_id, other_user) == ""

            live_client.user_name = live_client.session.headers[HTTP_HEADER_USR_ID] = other_user
            assert oai_obj.oaio_id not in live_client._load_server_objectz()

            live_client.user_name = live_client.session.headers[HTTP_HEADER_USR_ID] = oai_obj.username
            assert oai_obj.oaio_id in live_client._load_server_objectz()
            assert oai_obj.oaio_id in live_client.client_objectz

        finally:
            if oai_obj:
                assert live_client.unregister_object(oai_obj.oaio_id) == ""

    def test_unsubscribe_err(self, live_client):
        oai_obj = None
        try:
            oai_obj = live_client.register_object({})
            assert live_client.o_synchronize_with_server_if_online(), f"{live_client.error_message=}"
            assert oai_obj
            assert oai_obj.oaio_id in live_client._load_server_objectz()
            assert oai_obj.oaio_id in live_client.client_objectz

            err_msg = live_client.unsubscribe(oai_obj.oaio_id, oai_obj.username)  # ERR: creator has to unregister

            assert oai_obj.oaio_id in err_msg
            assert oai_obj.username in err_msg
            assert oai_obj.oaio_id in live_client._load_server_objectz()
            assert oai_obj.oaio_id in live_client.client_objectz

        finally:
            if oai_obj:
                assert live_client.unregister_object(oai_obj.oaio_id) == ""
            assert oai_obj.oaio_id not in live_client._load_server_objectz()
            assert oai_obj.oaio_id not in live_client.client_objectz

    def test_update_file(self, live_client):
        reg_stamp = now_stamp()
        fil_name = 'tst_fil.zzz'
        fil_content = b"test file \x00 binary content \x01\x02\x03\xff"
        local_object_count = len(live_client.client_objectz)
        reg_obj = live_client.register_file(fil_name, file_content=b"", stamp=reg_stamp)
        oaio_id = reg_obj.oaio_id
        assert live_client.o_synchronize_with_server_if_online(), f"{live_client.error_message=}"

        upd_stamp = now_stamp()
        oai_obj = live_client.update_file(oaio_id, file_name=fil_name, file_content=fil_content, stamp=upd_stamp)

        assert oai_obj is not None
        assert oai_obj.oaio_id == oaio_id
        dir_path = live_client._client_file_root(oaio_id, oai_obj.client_values)
        assert live_client.error_message == ""

        assert ROOT_VALUES_KEY not in oai_obj.client_values
        assert FILES_VALUES_KEY in oai_obj.client_values
        assert len(oai_obj.client_values[FILES_VALUES_KEY]) == 1
        assert oai_obj.client_values[FILES_VALUES_KEY][0] == fil_name
        assert os_path_isfile(os_path_join(dir_path, fil_name))
        assert oai_obj.server_values == oai_obj.client_values
        assert oai_obj.client_stamp == upd_stamp
        assert oai_obj.server_stamp == reg_stamp
        assert oai_obj.csh_access_right == CREATE_ACCESS_RIGHT

        assert len(live_client.client_objectz) == local_object_count + 1
        assert oaio_id in live_client.client_objectz
        assert live_client.client_objectz[oaio_id] == oai_obj

        assert _csh_file_content(live_client, oai_obj) == {fil_name: None}

        assert live_client.o_synchronize_with_server_if_online(), f"{live_client.error_message=}"

        assert _csh_file_content(live_client, oai_obj) == {fil_name: fil_content}

        assert _csh_file_remove(live_client, oai_obj) == ""
        assert live_client.unregister_object(oai_obj.oaio_id) == ""
        if os_path_isdir(dir_path):
            shutil.rmtree(dir_path)

    def test_update_folder_with_root(self, live_client):
        reg_stamp = now_stamp()
        folder_path = os_path_abspath(os_path_join(TESTS_FOLDER, 'fol_upd_root'))
        sub_dirs = ['sub_dir1', 'sub_dir3/sub_sub_dir']
        file_paths = ['tst_fil.tst',
                      os_path_join(sub_dirs[0], 'tst_sub1_fil.tst'),
                      os_path_join(sub_dirs[1], 'tst_sub3_sub_fil.tst'),
                      ]
        fil_content = b"test file \x00 binary content \x01\x02\x03\xff"

        local_object_count = len(live_client.client_objectz)
        csh_api = live_client._csh_api(live_client.csh_default_id)
        for fil_nam in file_paths:
            write_file(os_path_join(folder_path, fil_nam), fil_content + fil_nam.encode(), make_dirs=True)
        reg_obj = live_client.register_folder(folder_path, stamp=reg_stamp)
        assert live_client.o_synchronize_with_server_if_online(), f"{live_client.error_message=}"

        upd_stamp = now_stamp()
        for fil_nam in file_paths:
            write_file(os_path_join(folder_path, fil_nam), fil_nam.encode() + b"-updated")
        oai_obj = live_client.update_folder(reg_obj.oaio_id, root_path=folder_path, stamp=upd_stamp)

        oaio_id = oai_obj.oaio_id
        assert oai_obj is not None
        assert oai_obj.oaio_id
        assert live_client.error_message == ""
        assert oai_obj.csh_id
        assert ROOT_VALUES_KEY in oai_obj.client_values
        assert normalize(oai_obj.client_values[ROOT_VALUES_KEY]) == normalize(folder_path)
        assert FILES_VALUES_KEY in oai_obj.client_values
        assert len(oai_obj.client_values[FILES_VALUES_KEY]) == len(file_paths)
        svr_path = os_path_join(oaio_id, upd_stamp)
        for fil_nam in file_paths:
            assert fil_nam in oai_obj.client_values[FILES_VALUES_KEY]
            assert os_path_isfile(normalize(os_path_join(folder_path, fil_nam)))
            assert csh_api.deployed_file_content(os_path_join(svr_path, fil_nam)) is None
        assert oai_obj.client_stamp == upd_stamp
        assert oai_obj.server_stamp == reg_stamp
        assert oai_obj.server_values == oai_obj.client_values
        assert oai_obj.csh_access_right == CREATE_ACCESS_RIGHT
        assert len(live_client.client_objectz) == local_object_count + 1
        assert oaio_id in live_client.client_objectz
        assert live_client.client_objectz[oaio_id] == oai_obj

        assert live_client.o_synchronize_with_server_if_online(), f"{live_client.error_message=}"

        assert live_client.error_message == ""
        oai_obj = live_client.client_objectz.get(oai_obj.oaio_id)
        assert oai_obj
        assert oai_obj.oaio_id
        assert oai_obj.csh_id
        assert ROOT_VALUES_KEY in oai_obj.client_values
        assert normalize(oai_obj.client_values[ROOT_VALUES_KEY]) == normalize(folder_path)
        assert FILES_VALUES_KEY in oai_obj.client_values
        assert len(oai_obj.client_values[FILES_VALUES_KEY]) == len(file_paths)
        svr_path = os_path_join(oaio_id, upd_stamp)
        for fil_nam in file_paths:
            assert fil_nam in oai_obj.client_values[FILES_VALUES_KEY]
            assert os_path_isfile(normalize(os_path_join(folder_path, fil_nam)))
            assert csh_api.deployed_file_content(os_path_join(svr_path, fil_nam)) == fil_nam.encode() + b"-updated"
        assert oai_obj.client_stamp == upd_stamp
        assert oai_obj.server_stamp == upd_stamp, f"reg={reg_stamp} upd={upd_stamp}"
        assert oai_obj.server_values == oai_obj.client_values
        assert oai_obj.csh_access_right == CREATE_ACCESS_RIGHT
        assert len(live_client.client_objectz) == local_object_count + 1

        # cleanup
        if oaio_id:
            assert live_client.unregister_object(oaio_id, wipe_files=True) == ""
            assert len(live_client.client_objectz) == local_object_count
            err_msg = _csh_file_remove(live_client, oai_obj, file_count=len(file_paths))
            assert reg_stamp in err_msg
            assert oaio_id in err_msg
        if os_path_isdir(folder_path):
            shutil.rmtree(folder_path)

    def test_update_object(self, live_client):
        reg_stamp = now_stamp()
        reg_values = {'tst_str': 'tst_val', 'tst_int': 69}
        local_object_count = len(live_client.client_objectz)
        oai_obj = live_client.register_object(reg_values, stamp=reg_stamp)

        upd_values = {'tst_str': '_val', 'tst_int': 369}
        upd_stamp = now_stamp()
        oai_obj = live_client.update_object(oai_obj.oaio_id, upd_values, stamp=upd_stamp)
        assert live_client.error_message == ""
        assert oai_obj is not None

        assert oai_obj.oaio_id
        assert oai_obj.csh_id
        assert oai_obj.client_values == upd_values
        assert not oai_obj.server_values
        assert oai_obj.client_stamp == upd_stamp
        assert not oai_obj.server_stamp
        assert oai_obj.csh_access_right == CREATE_ACCESS_RIGHT
        assert len(live_client.client_objectz) == local_object_count + 1
        assert oai_obj.oaio_id in live_client.client_objectz
        assert live_client.client_objectz[oai_obj.oaio_id] == oai_obj

        assert live_client.o_synchronize_with_server_if_online(), f"{live_client.error_message=}"

        assert live_client.error_message == ""
        oai_obj = live_client.client_objectz.get(oai_obj.oaio_id)
        assert oai_obj
        assert oai_obj.oaio_id
        assert oai_obj.csh_id
        assert oai_obj.client_values == upd_values
        assert oai_obj.server_values == upd_values
        assert oai_obj.client_stamp == upd_stamp
        assert oai_obj.server_stamp == upd_stamp
        assert oai_obj.csh_access_right == CREATE_ACCESS_RIGHT
        assert len(live_client.client_objectz) == local_object_count + 1

        if oai_obj:
            assert live_client.unregister_object(oai_obj.oaio_id) == ""
            assert len(live_client.client_objectz) == local_object_count

    def test_upsert_subscription(self, live_client):
        oai_obj = None
        pid = 0
        try:
            oai_obj = live_client.register_object({})
            assert live_client.o_synchronize_with_server_if_online(), f"{live_client.error_message=}"

            assert oai_obj
            assert oai_obj.oaio_id in live_client._load_server_objectz()
            assert oai_obj.oaio_id in live_client.client_objectz
            assert oai_obj.username == live_client.user_name
            assert oai_obj.csh_access_right == CREATE_ACCESS_RIGHT

            pid = live_client.upsert_subscription(oai_obj.oaio_id, other_user, access_right=UPDATE_ACCESS_RIGHT)
            assert live_client.error_message == ""
            assert pid != 0

            # to allow check the subscription of other user and to mock a request, OAIO_USERNAME has to be supervisor
            live_client.user_name = live_client.session.headers[HTTP_HEADER_USR_ID] = other_user
            assert oai_obj.oaio_id in live_client._load_server_objectz(), f"{live_client.error_message=}"

        finally:
            if pid:
                live_client.user_name = live_client.session.headers[HTTP_HEADER_USR_ID] = oai_obj.username
                assert live_client.wipe_user(other_user) == ""
            if oai_obj:
                assert live_client.unregister_object(oai_obj.oaio_id) == ""

    def test_userz_access(self, live_client):
        users = live_client.userz_access("any-id")
        assert isinstance(users, list)
        assert len(users) >= 1
        assert isinstance(users[0], dict)
        assert 'username' in users[0]
        assert 'access_right' in users[0]

    def test_wipe_user(self, live_client):
        oai_obj = None
        try:
            oai_obj = live_client.register_object({})
            assert live_client.o_synchronize_with_server_if_online(), f"{live_client.error_message=}"
            assert live_client.upsert_subscription(oai_obj.oaio_id, other_user)  # create User/Userz recs for other user
            assert other_user in [_['username'] for _ in live_client.userz_access(oai_obj.oaio_id)]

            assert live_client.wipe_user(other_user) == ""
            assert other_user not in [_['username'] for _ in live_client.userz_access(oai_obj.oaio_id)]

        finally:
            if oai_obj:
                assert live_client.unregister_object(oai_obj.oaio_id) == ""

    def test_teardown(self):
        """ last live test method, to display teardown errors of live_client fixture (with session scope) separately """
